/* Nessus
 * Copyright (C) 1998 - 2001 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * In addition, as a special exception, Renaud Deraison
 * gives permission to link the code of this program with any
 * version of the OpenSSL library which is distributed under a
 * license identical to that listed in the included COPYING.OpenSSL
 * file, and distribute linked combinations including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * this file, you may extend this exception to your version of the
 * file, but you are not obligated to do so.  If you do not wish to
 * do so, delete this exception statement from your version.
 *
 */

#include "sussen.h"
#include "sussen-util.h"
#include "sussen-cert-druid.h"

#include <includes.h>
#include "nessus.h"
#include "globals.h"

#include "comm.h"
#include "auth.h"
#include "preferences.h"

# define NUM_ARGS 6

#ifndef USE_AF_INET
#   undef  NUM_ARGS
#   define NUM_ARGS 4
#endif


static int verify_callback(gint preverify_ok, X509_STORE_CTX *ctx) {	
	#if DEBUG_SSL > 1
		fprintf(stderr, "verify_callback> preverify_ok=%d\n", preverify_ok);
	#endif
	return preverify_ok;
}

/*
 * split a line "var=value" into two components
 * returns 0 if = was not found, 1 if line looks like "var=", 2 if OK
 */
static int split_var_val(const gchar *line, gchar *var, gchar *val) {

	const char *p;
	char *q;

	for (p = line, q = var; *p != '=' && *p != '\0' && *p != '\n'; p ++, q ++)
		*q = *p;
		*q = '\0';

	if (*p == '\0')
		return 0;
  
	for (q = val, p ++; *p != '\0' && *p != '\n'; p ++, q ++)
		*q = *p;
		*q = '\0';

	return q == line ? 1 : 2;
}

/* 
 * Returns -1 if error, 0 if hash not found, 1 if found
 */
static int get_server_cert_hash(const gchar *sname, guchar *hash) {

	gchar *fname;
	FILE *fp;
	gchar line[1024];
	gchar ho[1024], ha[1024];
	gint i, x;

	if ((fname = preferences_get_altname("cert")) == NULL)
		return -1;

	fp = fopen(fname, "r");
	efree(&fname);
	
	if (fp == NULL) {
		if (errno == ENOENT)
			return 0;
		else
			return -1;
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (split_var_val(line, ho, ha) == 2) {
			if (strcmp(ho, sname) == 0 && strlen(ha) == SHA_DIGEST_LENGTH * 2) {
				for (i = 0; i < SHA_DIGEST_LENGTH; i ++) {
					(void) sscanf(ha + 2 * i, "%2x", &x);
					hash[i] = x;
				}

				fclose(fp);
				return 1;
			}
		}
	}

	if (ferror(fp))
		return -1;
	else
		return 0;
}

static void print_hash(gchar *hash_str, const guchar *hash) {
	
	gint	i;

	for (i = 0; i < SHA_DIGEST_LENGTH; i ++)
		sprintf(hash_str + 2 * i, "%02x", hash[i]);
}

static int set_server_cert_hash(const gchar *sname, guchar *hash) {

	char	ho[1024], ha[1024];
	char	*fname = NULL;
	FILE	*fp1 = NULL, *fp2 = NULL;
	char	line[1024];
	int	x;
	int	found;

	if ((fname = preferences_get_altname("cert")) == NULL)
		return -1;
  
	if ((fp2 = tmpfile()) == NULL)
		goto error;
  
	fp1 = fopen(fname, "r");
	if (fp1 == NULL && errno != ENOENT)
		goto error;

	found = 0;
	if (fp1 != NULL) {
		while (fgets(line, sizeof(line), fp1) != NULL) {
			x = strlen(line);
	
			if (x > 0 && line[x - 1] != '\n') /* invalid line */
				continue;

			if (split_var_val(line, ho, ha) == 2) {
				if (strcmp(ho, sname) == 0) {
					if (found) /* multiple lines */
						continue;
					else {
						print_hash(ha, hash);
						sprintf(line, "%s=%s\n", ho, ha);
						found = 1;
					}
				}
			}

			if (fputs(line, fp2) < 0)
				goto error;
		}
	
		(void) fclose(fp1);
	}

	if (! found) {
		print_hash(ha, hash);
		sprintf(line, "%s=%s\n", sname, ha);

		if (fputs(line, fp2) < 0)
			goto error;
	}

	rewind(fp2);
	if ((fp1 = fopen(fname, "w")) == NULL)
		goto error;

	while (fgets(line, sizeof(line), fp2) != NULL)
		(void) fputs(line, fp1);

	if (ferror(fp1) || fclose(fp1) < 0)
		goto error;
	
	(void) fclose(fp2);		/* auto delete */
	efree(&fname);
	return 0;

	error:
		if (fp1 != NULL)
			fclose(fp1);
		if (fp2 != NULL)
			fclose(fp2);
		if (fname != NULL)
			efree(&fname);
		return -1;  
}


/*
 * connect_to_nessusd
 *
 * This function establishes the connection between
 * nessus and nessusd, logs in and reads the plugin
 * list from the server.
 *
 */

char *connect_to_nessusd (const char *hostname, int port, const char *login, const char *pass) {

	static SSL_CTX *ssl_ctx = NULL;
	static SSL_METHOD *ssl_mt = NULL;
	SSL *ssl = NULL;
	char *cert, *key, *client_ca, *trusted_ca, *ssl_ver;
	STACK_OF(X509_NAME)	*cert_names;

	int soc, soc2;
	int opt;

	init_globals();

	if(arg_get_type(Prefs, "nessusd_host")>=0)
		arg_set_value(Prefs, "nessusd_host", strlen(hostname), strdup(hostname));
	else
		arg_add_value(Prefs, "nessusd_host", ARG_STRING, strlen(hostname),
	strdup(hostname));
			
	if(arg_get_type(Prefs, "nessusd_user")>=0)
		arg_set_value(Prefs, "nessusd_user", strlen(login), strdup(login));
	else
		arg_add_value(Prefs, "nessusd_user", ARG_STRING, strlen(login),
	strdup(login));		

	soc = open_sock_tcp_hn(hostname, port);
  
  if(soc<0) {
	struct in_addr a = nn_resolve(hostname);
	
	if(a.s_addr == INADDR_NONE) 
		return("Host not found !");
	else
		return("Could not open a connection to the remote host");
	}
	
	opt = 1;	
	setsockopt(soc, SOL_SOCKET, SO_KEEPALIVE, &opt, sizeof(opt));
	
	#define SSL_VER_DEF_NAME	"TLSv1"
	#define SSL_VER_DEF_METH	TLSv1_client_method
	ssl_ver = arg_get_value(Prefs, "ssl_version");

	if (ssl_ver == NULL || *ssl_ver == '\0')
		ssl_ver = SSL_VER_DEF_NAME;

	if (strcasecmp(ssl_ver, "NONE") != 0) {
		if(nessus_SSL_init(NULL) < 0) {
			return("Could not initialize the OpenSSL library !\n\
					Please launch nessus-mkrand(1) first !");
		}

		if (ssl_mt == NULL) {
			if (strcasecmp(ssl_ver, "SSLv2") == 0)
				ssl_mt = SSLv2_client_method();
			else if (strcasecmp(ssl_ver, "SSLv3") == 0)
				ssl_mt = SSLv3_client_method();
			else if (strcasecmp(ssl_ver, "SSLv23") == 0)
				ssl_mt = SSLv23_client_method();
			else if (strcasecmp(ssl_ver, "TLSv1") == 0)
				ssl_mt = TLSv1_client_method();
			else {
				fprintf(stderr, "Unknown SSL version \"%s\"\nSwitching to default " SSL_VER_DEF_NAME "\n", ssl_ver);
				ssl_ver = SSL_VER_DEF_NAME;
				ssl_mt = SSL_VER_DEF_METH();
			}
      
			if (ssl_mt == NULL) {
				char	s[32];
				sprintf(s, "%s_client_method", ssl_ver);
				sslerror(s);
				return "SSL error";
			}
		}

		if (ssl_ctx == NULL)
			if ((ssl_ctx = SSL_CTX_new(ssl_mt)) == NULL) {
				sslerror("SSL_CTX_new");
				return "SSL error";
			}

		if (SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL) < 0)
			sslerror("SSL_CTX_set_options(SSL_OP_ALL)");

		if ((ssl = SSL_new(ssl_ctx)) == NULL) {
			sslerror("SSL_new");
			return "SSL_error";
		}

		cert = arg_get_value(Prefs, "cert_file");
		key = arg_get_value(Prefs, "key_file");
		client_ca = arg_get_value(Prefs, "client_ca");

		if (pass != NULL && key != NULL) {		
			nessus_install_passwd_cb(ssl_ctx, (char *)pass);
		}

		if (cert != NULL)
			SSL_use_certificate_file(ssl, cert, SSL_FILETYPE_PEM);
		if (key != NULL)
			SSL_use_PrivateKey_file(ssl, key, SSL_FILETYPE_PEM);

		if (client_ca != NULL) {
			cert_names = SSL_load_client_CA_file(client_ca);

			if (cert_names != NULL)
				SSL_CTX_set_client_CA_list(ssl_ctx, cert_names);
			else
				sslerror("SSL_load_client_CA_file");
		}

		if (! SSL_set_fd(ssl, soc)) {
			sslerror("SSL_set_fd");
			return "SSL error";
		}

		if (SSL_connect(ssl) <= 0) {
			sslerror("SSL_connect");
			return "SSL error";
		}
		
		X509 *x509_cert = SSL_get_peer_certificate (ssl);
		char stored_hash [SHA_DIGEST_LENGTH];
		
		if (get_server_cert_hash(hostname, stored_hash) <= 0)
			bzero (stored_hash, sizeof(stored_hash));
		
		if (x509_cert == NULL) {
			sslerror ("SSL_get_peer_certificate");
			g_print ("SSL Error: cannot get server certificate\n");
		}
		
		X509_check_purpose (x509_cert, -1, 0);
		
		if (memcmp(x509_cert->sha1_hash, stored_hash, SHA_DIGEST_LENGTH) != 0) {
			int value;
			
			value = show_cert_druid (ssl);		
			switch (value) {
				
				case CERT_DRUID_ACCEPT:
					if (set_server_cert_hash(hostname, x509_cert->sha1_hash) < 0)
						return "Could not save server certificate";
					break;
				
				case CERT_DRUID_VERIFY:
					trusted_ca = arg_get_value(Prefs, "trusted_ca");

					if (trusted_ca == NULL) {
						return ("no trusted_ca\n");
					} else {
						SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_PEER, verify_callback);
				
						if (SSL_CTX_set_default_verify_paths(ssl_ctx) <= 0)
							return ("SSL_CTX_set_default_verify_paths");
						if (! SSL_CTX_load_verify_locations(ssl_ctx, trusted_ca, NULL))
							return ("SSL_CTX_load_verify_locations");
					}
					break;

				case CERT_DRUID_REJECT:
					return "invalid certificate";
					break;

				default:
					break;
			}
		}
	} /* ssl_ver != "NONE" */

	if ((soc2 = nessus_register_connection(soc, ssl)) <0) {
		shutdown(soc, 2);
		return "Could not register the connection";
	}
	soc = soc2;
  
	GlobalSocket = soc;
	ArgSock = emalloc(sizeof(struct arglist));
	arg_add_value(ArgSock, "global_socket", ARG_INT, -1, (void *)GlobalSocket);

	if(comm_init(soc,PROTO_NAME) || (auth_login(login, pass))) {
		close_stream_connection(GlobalSocket);
		return("Remote host is not using the good version of the Nessus communication protocol (1.2) or is tcpwrapped");
	}

	if(comm_get_plugins())
		return("Login failed");
	
	if(!First_time) {
		comm_get_preferences(Prefs);
		comm_get_rules(Prefs);	
		comm_get_dependencies();
	} else {
		/*
		 * Ignore the server preferences if we already logged in
		 */
		struct arglist * devnull = emalloc(sizeof(*devnull));
		comm_get_preferences(devnull);
		comm_get_rules(devnull);
		arg_free(devnull);
	}
	
	if(comm_server_restores_sessions(Prefs)) {
		Sessions = comm_get_sessions();
	}
	
	set_target_defaults ();
	set_widgets_sensitive (TRUE,FALSE);

	GtkWidget *w;
	
	w = glade_xml_get_widget (mainwindow_xml,"connect_button");
	gtk_widget_set_sensitive (w, FALSE);
	w = glade_xml_get_widget (mainwindow_xml,"connect1");
	gtk_widget_set_sensitive (w, FALSE);	
	w = glade_xml_get_widget (mainwindow_xml,"stop_scan_button");
	gtk_widget_set_sensitive (w, FALSE);
	w = glade_xml_get_widget (mainwindow_xml,"stop_scan1");
	gtk_widget_set_sensitive (w, FALSE);	
	w = glade_xml_get_widget (mainwindow_xml,"stop_scanning_host_button");
	gtk_widget_set_sensitive (w, FALSE);


	return(NULL);
}

/*
 * init_globals
 *
 * initializes two main global variables : Plugins and
 * Scanners
 *
 */

void init_globals(void) {
	if (!Plugins)
		Plugins = emalloc(sizeof(struct arglist));

	if (!Scanners)
		Scanners = emalloc(sizeof(struct arglist));
}
